#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>




#define PI 3.14159265358

float radians(float degree){
    return ((degree*PI)/180.0);
}

float degrees(float radian){
    return ((radian*180.0)/PI);
}

#include "Servo.h"
#include "EEPROM.h"
#include "Device_I2C.h"
#include "Wire.h"
#include "GlobalDefined.h"
#include "AeroQuad.h"
#include "PID.h"
#include "AQMath.h"
#include "FourtOrderFilter.h"

volatile long int AbsoluteTime = 0;   


//********************************************************
//********************************************************
//********* PLATFORM SPECIFIC SECTION ********************
//********************************************************
//********************************************************


// Gyroscope declaration
#define ITG3200_ADDRESS_ALTERNATE
#include "Gyroscope_ITG3200.h"

// Accelerometer declaration
#include "Accelerometer_ADXL345.h"

// Receiver declaration
#define RECEIVER_328P

// Motor declaration
#define MOTOR_PWM



/**
 * Put AeroQuad_Mini specific intialization need here
 */
void initPlatform() {
    
    DDRD |= (1<<PIN2);        //pinMode(LED_Red, OUTPUT);
    PORTD |= (1<<PIN2);
    //PORTD &= ~(1<<PIN2);      //digitalWrite(LED_Red, LOW);
    //pinMode(LED_Yellow, OUTPUT);
    //digitalWrite(LED_Yellow, LOW);
    
    
    Wire.begin();
    TWBR = 12;
}

/**
 * Measure critical sensors
 */
void measureCriticalSensors() {
    measureAccelSum();
    measureGyroSum();
}


//********************************************************
//********************************************************
//********* HARDWARE GENERALIZATION SECTION **************
//********************************************************
//********************************************************

//********************************************************
//****************** KINEMATICS DECLARATION **************
//********************************************************
#include "Kinematics.h"
#include "Kinematics_ARG.h"


//********************************************************
//******************** RECEIVER DECLARATION **************
//*******************************************************/
#include "Receiver_RemotePC.h"



//********************************************************
//********************** MOTORS DECLARATION **************
//********************************************************
#include "Motors_PWM.h"

//********************************************************
//******** FLIGHT CONFIGURATION DECLARATION **************
//********************************************************
#include "FlightControlQuadX.h"

// Include this last as it contains objects from above declarations
#include "AltitudeControlProcessor.h"
#include "FlightControlProcessor.h"
#include "FlightCommandProcessor.h"
#include "HeadingHoldProcessor.h"
#include "DataStorage.h"
#include "SerialCom.h"
#include "HardwareSerial.h"
//#include "uart.h"
#include <string.h>
#include <stdlib.h>


/**
 * Main setup function, called one time at bootup
 * initalize all system and sub system of the 
 * 
 */


char receiver_buffer[10]="0";
int receiver_index=0;

HardwareSerial Serial;

void getReceiverValues(void);

void setup() {
    
    begin(Serial,2400);
    write(Serial, "Initializing Setup Parameters....\n");
    
    TIMSK0=2;
    
    OCR0A=250;

    TCCR0A=2;
    TCCR0B=3;
    sei();

    
    
    
#ifdef CHANGE_YAW_DIRECTION
    YAW_DIRECTION = -1;
#endif
    
    // Read user values from EEPROM
    readEEPROM(); // defined in DataStorage.h
    if (readFloat(SOFTWARE_VERSION_ADR) != SOFTWARE_VERSION) { // If we detect the wrong soft version, we init all parameters
        initializeEEPROM();
        writeEEPROM();
    }
    
    initPlatform();
    
    // Configure motors
    initializeMotors(FOUR_Motors);
    // Initialize max/min values for all motors
    for (uint8_t motor = 0; motor < LASTMOTOR; motor++) {
        motorMinCommand[motor] = minArmedThrottle;
        motorMaxCommand[motor] = MAXCOMMAND;
    }
    
    // Setup receiver pins for pin change interrupts
    initializeReceiver(LASTCHANNEL);
    initReceiverFromEEPROM();
    
    // Initialize sensors
    // If sensors have a common initialization routine
   
    initializeGyro(); // defined in Gyro.h
    initializeAccel(); // defined in Accel.h
    initSensorsZeroFromEEPROM();
    
    // Calibrate sensors
    calibrateGyro();
    computeAccelBias();
    zeroIntegralError();
    
    // Flight angle estimation

    initializeKinematics(1.0, 0.0);  
    
    PID[ATTITUDE_XAXIS_PID_IDX].windupGuard = 0.375;
    PID[ATTITUDE_YAXIS_PID_IDX].windupGuard = 0.375;
    
        
    setupFourthOrder();
    
    previousTime = AbsoluteTime;
    //PORTD &= ~(1<<PIN2);   //digitalWrite(LED_Green, HIGH);     Hardware reversed
    safetyCheck = 0;
}


/*******************************************************************
 // tasks (microseconds of interval)
 ReadGyro        readGyro      (   5000); // 200hz
 ReadAccel       readAccel     (   5000); // 200hz
 FlightControls  flightControls(  10000); // 100hz
 ReadReceiver    readReceiver  (  20000); //  50hz
 *******************************************************************/


int main(void){
    setup();
     
 while (1) {
    currentTime = AbsoluteTime;
    deltaTime = currentTime - previousTime;
    
    measureCriticalSensors();
    
    // Main scheduler loop set for 100hz
    if (deltaTime >= 10) {
        
        frameCounter++;
        
        // ================================================================
        // 100hz task loop
        // ================================================================
        if (frameCounter % TASK_100HZ == 0) {  //  100 Hz tasks
            
            G_Dt = (currentTime - hundredHZpreviousTime) / 1000000.0;
            hundredHZpreviousTime = currentTime;
            
            evaluateMetersPerSec();
            evaluateGyroRate();
            
         
            
            for (int axis = XAXIS; axis <= ZAXIS; axis++) {
                filteredAccel[axis] = computeFourthOrder(meterPerSecSec[axis], &fourthOrder[axis]);
            }
            
            // ****************** Calculate Absolute Angle *****************

            
            calculateKinematics(gyroRate[XAXIS],
                                 gyroRate[YAXIS],
                                gyroRate[ZAXIS],
                                filteredAccel[XAXIS],
                                filteredAccel[YAXIS],
                                filteredAccel[ZAXIS],
                                0.0,
                                0.0,
                                0.0,
                                G_Dt);
                        
            processFlightControl();
        }
        
        // ================================================================
        // 50hz task loop
        // ================================================================
        if (frameCounter % TASK_50HZ == 0) {  //  50 Hz tasks
            
            G_Dt = (currentTime - fiftyHZpreviousTime) / 1000000.0;
            fiftyHZpreviousTime = currentTime;
            
            // Reads external pilot commands and performs functions based on stick configuration
            getReceiverValues();
            readPilotCommands(); 
            
        }
        
        // ================================================================
        // 10hz task loop
        // ================================================================
        if (frameCounter % TASK_10HZ == 0) {  //   10 Hz tasks
            
            G_Dt = (currentTime - tenHZpreviousTime) / 1000000.0;
            tenHZpreviousTime = currentTime;
            
            
           
            
            
        }
        
        previousTime = currentTime;
    }
    if (frameCounter >= 100) {
        frameCounter = 0;
    }
}

}

void getReceiverValues(void){
    char newChar;
    while (Serial.available() > 0) {
        newChar = Serial.read();
        if(newChar >= '0' && newChar <= '9'){
            receiver_buffer[receiver_index++] = newChar;
        }else if(newChar >= 'A' && newChar <= 'z'){
            switch(newChar){
                case 'Y': receiver_buffer[receiver_index]='\0';receiver_index=0;receiverBuffered[XAXIS] = atoi(receiver_buffer);break;
                case 'P': receiver_buffer[receiver_index]='\0';receiver_index=0;receiverBuffered[YAXIS] = atoi(receiver_buffer);break;
                case 'R': receiver_buffer[receiver_index]='\0';receiver_index=0;receiverBuffered[ZAXIS] = atoi(receiver_buffer);break;
                case 'T': receiver_buffer[receiver_index]='\0';receiver_index=0;receiverBuffered[THROTTLE] = atoi(receiver_buffer);break;
                default: Serial.write("Unkwon Value\n\r");
                    
            }
        }
        
    }
    Serial.write("\n\rX: ");
    Serial.printNumber(receiverCommand[XAXIS], 10);
    Serial.write("  Y: ");
    Serial.printNumber(receiverCommand[YAXIS], 10);
    Serial.write("  Z: ");
    Serial.printNumber(receiverCommand[ZAXIS], 10);
    Serial.write("  T: ");
    Serial.printNumber(receiverCommand[THROTTLE], 10);

}

